package org.zdevra.sparrow.actor;

import org.zdevra.sparrow.*;
import org.zdevra.sparrow.channel.MessageChannel;
import org.zdevra.sparrow.Nameable;
import org.zdevra.sparrow.mailbox.FifoBlockingMailbox;
import org.zdevra.sparrow.mailbox.Mailbox;
import org.zdevra.sparrow.message.Message;

import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Actor is object with async. message queue called {@link Mailbox mailbox} and execution procedure for processing
 * messages from that queue. One actor refers to one thread. Actor's lifecycle is managed by {@link SparrowEngine engine}
 * where actor is placed in. That means also the actor should be constructed via engine's builder methods.
 *
 * <br><br>
 * All entities/patterns like aggregators, smart proxy etc. are based on this actor.

 * @author Zdenko Vrabel (zdenko.vrabel@celum.com)
 */
public abstract class AbstractActor implements SparrowLifecycle, MessageChannel, Runnable, Nameable
{
	//-------------------------------------------------------------------------------------------------------------------
	// members
	//-------------------------------------------------------------------------------------------------------------------

	/** log instance */
	private static final Logger logger = Logger.getLogger(AbstractActor.class.getName());

	/** mailbox implementation for actor. Messages are stored primary in this mailbox */
	private final Mailbox mailbox;

	/** handler for all uncaught exceptions thrown in actor's handler */
	private final ActorExceptionHandler exceptionHandler;

	/** actor's thread. Each actor has his own processing thread */
	private volatile Thread actorThread;

	/** actor's unique name */
	private volatile String name;


	//-------------------------------------------------------------------------------------------------------------------
	// constructors
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Constructor with default mailbox and exception handler
	 */
	public AbstractActor()
	{
		this(null, null);
	}


	/**
	 * Constructor
	 *
	 * @param mailbox
	 *           mailbox which the actor uses. If it's null, default mailbox is used (Fifo with capacity 100)
	 *
	 * @param exceptionHandler
	 *           all exceptions are handled inside this handler
	 */
	public AbstractActor(Mailbox mailbox, ActorExceptionHandler exceptionHandler)
	{
		if (mailbox == null) {
			this.mailbox = new FifoBlockingMailbox(100);
		} else {
			this.mailbox = mailbox;
		}

		if (exceptionHandler != null) {
			this.exceptionHandler = exceptionHandler;
		} else {
			this.exceptionHandler = new DummyExceptionHandler();
		}
	}


	//-------------------------------------------------------------------------------------------------------------------
	// abstract methods
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Here you will implement message processing procedure.
	 *
	 * @param msg
	 *           received message
	 */
	protected abstract void onReceive(Message<?> msg);


	//-------------------------------------------------------------------------------------------------------------------
	// impl. methods
	//-------------------------------------------------------------------------------------------------------------------


	@Override
	public String toString()
	{
		return "Actor(" + name + ")";
	}


	/**
	 * {@inheritDoc}
	 * @param name
	 */
	@Override
	public void setName(String name)
	{
		this.name = name;
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public String getName()
	{
		return name;
	}


	@Override
	public void onInit(SparrowEngine engine)
	{
		//empty impl.
	}


	/**
	 * Send message into actor's mailbox.
	 *
	 * {@inheritDoc}
	 */
	@Override
	public boolean send(Message<?> message)
	{
		return mailbox.offer(message);
	}


	/**
	 * Send message into actor's mailbox.
	 *
	 * {@inheritDoc}
	 */
	@Override
	public boolean send(Message<?> message, long timeout)
	{
		return mailbox.offer(message, timeout);
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public void onStart()
	{
		logger.log(Level.INFO, "Start actor:" + getName());
		if (actorThread == null) {
			actorThread = SparrowThreadFactory.createThread(this, name);
			actorThread.start();
		}
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public void onStop() throws InterruptedException
	{
		logger.log(Level.INFO, "Stop actor:" + getName());
		if (actorThread != null && !actorThread.isInterrupted()) {
			actorThread.interrupt();
		}
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public final void run()
	{
		while(!Thread.currentThread().isInterrupted())
		{
			Message<?> message = mailbox.take();
			receive(message);
		}
	}


	/**
	 * receiving of message
	 * @param msg
	 */
	private void receive(Message<?> msg)
	{
		try {
			if (msg != null) {
				logger.log(Level.FINEST, "received message " + msg.getHeader().getId());
				onReceive(msg);
			}
		} catch (Throwable e) {
			exceptionHandler.uncaughtException(Thread.currentThread(), e);
		}
	}

}
