/*****************************************************************************
 * Copyright 2012 Zdenko Vrabel
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package org.zdevra.sparrow.aggregator;

import org.zdevra.sparrow.message.Message;
import org.zdevra.sparrow.actor.OutputActorHandler;

import java.util.*;

/**
 * AggregatorActorHandler is oposite of the splitter and aggregate messages by correlation ID.
 * @author Zdenko Vrabel (zdenko.vrabel@celum.com)
 */
class AggregatorActorHandler extends OutputActorHandler<Object>
{

	//-------------------------------------------------------------------------------------------------------------------
	// members
	//-------------------------------------------------------------------------------------------------------------------

	/** map of all messages collected in message groups. Each message group is identified by UUID */
	private final Map<UUID, MessageGroup> messageGroups;

	private final long timeout;

	private final AggregatorStrategy aggregatorStrategy;

	private final ReleaseStrategy releaseStrategy;

	//-------------------------------------------------------------------------------------------------------------------
	// constructor
	//-------------------------------------------------------------------------------------------------------------------


	/**
	 * Constructor
	 */
	protected AggregatorActorHandler(AggregatorStrategy aggregatorStrategy, ReleaseStrategy releaseStrategy, long timeout)
	{
		this.messageGroups = new HashMap<UUID, MessageGroup>();
		this.aggregatorStrategy = aggregatorStrategy;
		this.releaseStrategy = releaseStrategy;
		this.timeout = timeout;
	}


	//-------------------------------------------------------------------------------------------------------------------
	// methods
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Aggregation procedure
	 *
	 * {@inheritDoc}
	 */
	@Override
	public final void act(Message<Object> msg)
	{
		//get message group
		UUID currentCorrelationId = msg.getHeader().getCorrelationId();
		MessageGroup msgGroup = messageGroups.get(currentCorrelationId);

		//if there is no existing group
		if (msgGroup == null) {
			msgGroup =
					  new MessageGroup(
								 msg.getHeader().getCorrelationId(),
								 msg.getHeader().getSequenceSize(),
								 timeout );

			messageGroups.put(currentCorrelationId, msgGroup);
		}

		//add message to group and do completion if group is complete
		if (msgGroup.addAndCheckCompletion(msg)) {
			Message<?> aggregatedMsg = aggregatorStrategy.aggregate(msgGroup.getMessages());
			if (aggregatedMsg != null) {
				output().send(aggregatedMsg);
				messageGroups.remove(currentCorrelationId);
			}
		}

		//releasing of expired groups will be performed
		releaseExpiredGroups();
	}


	/**
	 * method release all these groups,for that timeout limit expires
	 */
	private void releaseExpiredGroups()
	{
		//collect all old groups
		List<MessageGroup> expired = new LinkedList<MessageGroup>();
		for(Map.Entry<UUID, MessageGroup> entry : messageGroups.entrySet()) {
			if (entry.getValue().isExpired()) {
				expired.add(entry.getValue());
			}
		}

		//remove all collected groups
		for (MessageGroup group : expired)
		{
			if (releaseStrategy != null) {
				releaseStrategy.release(group.getMessages());
			}
			messageGroups.remove(group.getCorrelationId());
		}
	}

}
