/*****************************************************************************
 * Copyright 2012 Zdenko Vrabel
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package org.zdevra.sparrow.aggregator;

import org.zdevra.sparrow.message.Message;

import java.util.LinkedList;
import java.util.List;
import java.util.UUID;

/**
 * Represents message group for {@link AggregatorActorHandler}'s purpose
 *
 * @author Zdenko Vrabel (zdenko.vrabel@celum.com)
 */
class MessageGroup
{
	//-------------------------------------------------------------------------------------------------------------------
	// members
	//-------------------------------------------------------------------------------------------------------------------

	/** correlation ID that is key for group */
	private final UUID correlationId;

	/** timestamp when the group was created */
	private final long timestamp;

	/** timeout for checking whether group is old */
	private final long expirationtime;

	/** how many messages we're waiting for */
	private final int sequenceSize;

	/** group of messages */
	private final List<Message<?>> messages;

	//-------------------------------------------------------------------------------------------------------------------
	// constructor & methods
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Constructor
	 */
	MessageGroup(UUID correlationId, int sequenceSize, long timeout)
	{
		this.messages = new LinkedList<Message<?>>();
		this.timestamp = System.currentTimeMillis();
		this.expirationtime = this.timestamp + timeout;
		this.sequenceSize = sequenceSize;
		this.correlationId = correlationId;
	}


	/**
	 * check whether message group is too old and should be released from stack
	 *
	 * @return true if it's old otherwise false
	 */
	public boolean isExpired()
	{
		if (System.currentTimeMillis() > expirationtime) {
			return true;
		}
		return false;
	}


	/**
	 * Adds message into group and return boolean that tell you
	 * whether group is complete.
	 *
	 * @param msg
	 *
	 * @return true if group is complete, otherwise false
	 */
	public boolean addAndCheckCompletion(Message<?> msg)
	{
		messages.add(msg);
		if (messages.size() >= sequenceSize) {
			return true;
		}
		return false;
	}


	/**
	 * returns list of messages
	 */
	public List<Message<?>> getMessages()
	{
		return this.messages;
	}


	/**
	 * returns correlation ID for group
	 */
	public UUID getCorrelationId()
	{
		return correlationId;
	}
}
