/*****************************************************************************
 * Copyright 2012 Zdenko Vrabel
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package org.zdevra.sparrow.channel;

import org.zdevra.sparrow.Nameable;
import org.zdevra.sparrow.message.Message;

import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

/**
 * Base channel of publish-subscribe pattern. The message dispatching is happening in thread from which
 * was the sending method called. All subclasses implements dispatch algorithm.
 *
 * @author Zdenko Vrabel (vrabel.zdenko@gmail.com)
 */
public abstract class PublishSubscribeChannel implements SubscribableChannel, Nameable
{
	//-------------------------------------------------------------------------------------------------------------------
	// members
	//-------------------------------------------------------------------------------------------------------------------

	/** list of all registered subscribers */
	private volatile List<MessageHandler> subscribers;

	/** channel's name */
	private volatile String name;


	//-------------------------------------------------------------------------------------------------------------------
	// constructor
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Constructor
	 */
	public PublishSubscribeChannel()
	{
		subscribers = Collections.synchronizedList(new LinkedList<MessageHandler>());
	}


	//-------------------------------------------------------------------------------------------------------------------
	// methods
	//-------------------------------------------------------------------------------------------------------------------


	/**
	 * {@inheritDoc}
	 */
	@Override
	public void setName(String name)
	{
		this.name = name;
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public String getName()
	{
		return this.name;
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public void subscribe(MessageHandler handler)
	{
		subscribers.add(handler);
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean send(Message<?> message)
	{
		return dispatch(message, this.subscribers);
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean send(Message<?> message, long timeout)
	{
		return dispatch(message, this.subscribers);
	}


	/**
	 * Delivers message to all registered subscribers.
	 *
	 * @param msg
	 *
	 * @return true if all subscribers received message, false if one of the subscriber's receiving fail.
	 */
	protected abstract boolean dispatch(Message<?> msg, List<MessageHandler> subscribers);
}
