/*****************************************************************************
 * Copyright 2012 Zdenko Vrabel
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package org.zdevra.sparrow.channel;

import org.zdevra.sparrow.Nameable;
import org.zdevra.sparrow.message.Message;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

/**
 * This channel receive messages and store it into inner queue.
 *
 * @author Zdenko Vrabel (zdenko.vrabel@celum.com)
 */
public class QueueChannel implements MessageChannel, PollableChannel, Nameable
{
	//-------------------------------------------------------------------------------------------------------------------
	// members
	//-------------------------------------------------------------------------------------------------------------------

	/** queue for channel with some capacity */
	private final BlockingQueue<Message<?>> queue;

	/** channel name */
	private volatile String name;


	//-------------------------------------------------------------------------------------------------------------------
	// constructor
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Constructor
	 */
	public QueueChannel(int capacity)
	{
		this.queue = new LinkedBlockingQueue<Message<?>>(capacity);
	}


	//-------------------------------------------------------------------------------------------------------------------
	// methods
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * returns type and name of channel
	 */
	@Override
	public String toString()
	{
		return "QueueChannel(" + getName() + ")";
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void setName(String name)
	{
		this.name = name;
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public String getName()
	{
		return this.name;
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public Message<?> receive()
	{
		return receive(-1);
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public Message<?> receive(long timeout)
	{
		try {
			if (timeout > 0) {
				return queue.poll(timeout, TimeUnit.MILLISECONDS);
			} else if (timeout == 0) {
				return queue.poll();
			} else {
				return queue.take();
			}
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			return null;
		}
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean send(Message<?> message)
	{
		return send(message, -1);
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean send(Message<?> message, long timeout)
	{
		try {
			if (timeout > 0) {
				return queue.offer(message, timeout, TimeUnit.MILLISECONDS);
			} else if (timeout == 0) {
				return queue.offer(message);
			} else {
				queue.put(message);
				return true;
			}
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			return false;
		}
	}
}
