/*****************************************************************************
 * Copyright 2012 Zdenko Vrabel
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package org.zdevra.sparrow.filter;

import org.zdevra.sparrow.DiscardStrategy;
import org.zdevra.sparrow.actor.OutputActorHandler;
import org.zdevra.sparrow.message.Message;

import java.util.Collections;
import java.util.List;

/**
 * Filter implementation of {@link org.zdevra.sparrow.actor.ActorHandler} that does filtering based on
 * {@link FilterRule rules} with AND operator.
 *
 * @author Zdenko Vrabel (zdenko.vrabel@celum.com)
 */
class FilterActorHandler extends OutputActorHandler
{
	//-------------------------------------------------------------------------------------------------------------------
	// members
	//-------------------------------------------------------------------------------------------------------------------

	/** filter rules. Message have to be accepted by all filter rules */
	private final List<FilterRule> rules;

	/** discard strategy see {@link DiscardStrategy} */
	private final DiscardStrategy discard;


	//-------------------------------------------------------------------------------------------------------------------
	// constructor
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Constructor
	 * @param rules
	 */
	FilterActorHandler(List<FilterRule> rules)
	{
		this(rules, null);
	}


	/**
	 * Constructor
	 *
	 * @param rules
	 * @param discardStrategy
	 */
	FilterActorHandler(List<FilterRule> rules, DiscardStrategy discardStrategy)
	{
		this.rules = Collections.unmodifiableList(rules);
		this.discard = discardStrategy;
	}



	//-------------------------------------------------------------------------------------------------------------------
	// methods
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Method apply all filter {@link FilterRule rules} on message. When one of the rules reject the message,
	 * message will be refused at all.
	 *
	 * @param msg
	 */
	@Override
	public void act(Message msg)
	{
		for (FilterRule rule : rules) {
			boolean res = rule.accept(msg);
			if (!res) {
				discard.discard(msg);
				return;
			}
		}

		output().send(msg);
	}
}
