/*****************************************************************************
 * Copyright 2012 Zdenko Vrabel
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package org.zdevra.sparrow.mailbox;

import org.zdevra.sparrow.message.Message;

import java.util.concurrent.*;

/**
 * Basic FIFO queue implementation of mailbox. This mailbox change the InputOutputActor's send() behaviour to non-blocking. It
 * means the message will be refused if mailbox is full.
 *
 * @author Zdenko Vrabel (vrabel.zdenko@gmail.com)
 */
public class FifoMailbox implements Mailbox
{
	//-------------------------------------------------------------------------------------------------------------------
	// members
	//-------------------------------------------------------------------------------------------------------------------

	/** queue for mailbox messages */
	private final BlockingQueue<Message<?>> queue;

	//-------------------------------------------------------------------------------------------------------------------
	// constructors
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Constructor
	 */
	public FifoMailbox(int queueCapacity)
	{
		this.queue = new ArrayBlockingQueue<Message<?>>(queueCapacity);
	}

	//-------------------------------------------------------------------------------------------------------------------
	// impl. methods
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Message<?> take()
	{
		try {
			return queue.take();
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			return null;
		}
	}


	/**
	 * Method is non-blocking what means the actor will refuse the message if mailbox is full.
	 * {@inheritDoc}
	 */
	@Override
	public boolean offer(Message<?> msg)
	{
		return offer(msg, 0);
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public boolean offer(Message<?> msg, long timeout)
	{
		try {
			if (timeout > 0) {
				return queue.offer(msg, timeout, TimeUnit.MILLISECONDS);
			} else if (timeout == 0) {
				return queue.offer(msg);
			} else {
				queue.put(msg);
				return true;
			}
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			return false;
		}
	}
}
