/*****************************************************************************
 * Copyright 2012 Zdenko Vrabel
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package org.zdevra.sparrow.message;

import java.io.Serializable;
import java.util.Collections;
import java.util.Map;

/**
 * Generic message with payload T and additional data in message's header. Each message has his payload, and several
 * mandatory fields in header like unique ID, timestamp when the messages was created and expiration time when
 * is message invalid.
 *
 * @see MessageHeader
 * @author Zdenko Vrabel (vrabel.zdenko@gmail.com)
 */
public class Message<T> implements Serializable
{
	//-------------------------------------------------------------------------------------------------------------------
	// members
	//-------------------------------------------------------------------------------------------------------------------

	/** message's payload */
	private final T payload;

	/** message header. */
	private final MessageHeader header;

	//------------------------------------------------------------------------------------------------------------------
	// constructors
	//------------------------------------------------------------------------------------------------------------------

	/**
	 * Constructor
	 *
	 * @param payload
	 */
	public Message(T payload)
	{
		this(payload, Collections.<String, Object>emptyMap());
	}


	/**
	 * Constructor
	 *
	 * @param payload
	 * @param header
	 */
	public Message(T payload, MessageHeader header)
	{
		this(payload, header.getMap());
	}


	/**
	 * Constructor
	 *
	 * @param payload
	 * @param header
	 */
	public Message(T payload, Map<String, Object> header)
	{
		this.payload = payload;
		this.header = new MessageHeader(header);
	}


	//------------------------------------------------------------------------------------------------------------------
	// impl. methods
	//------------------------------------------------------------------------------------------------------------------

	@Override
	public String toString()
	{
		String out =  this.getClass().getCanonicalName() + "{ ";

		if (payload != null) {
			out += " payload=" + payload.toString();
		} else {
			out += ", payload=null";
		}

		out += ", header=" + header.toString();
		out += " }";
		return out;
	}

	//------------------------------------------------------------------------------------------------------------------
	// methods
	//------------------------------------------------------------------------------------------------------------------


	/**
	 * Return payload as type A
	 *
	 * @param <A>
	 *
	 * @return
	 */
	public <A> A getPayloadAs(Class<A> clazz)
	{
		return (A) getPayload();
	}


	/**
	 * Method returns the payload
	 * @return payload
	 */
	public T getPayload()
	{
		return payload;
	}


	/**
	 * Method returns header
	 * @return message's header
	 */
	public MessageHeader getHeader()
	{
		return this.header;
	}


	/**
	 * Cast message with payload's type to another one
	 * @param <A>
	 * @return
	 */
	public <A> Message<A> castTo()
	{
		Message<?> m = this;
		return (Message<A>) m;
	}
}