/*****************************************************************************
 * Copyright 2012 Zdenko Vrabel
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package org.zdevra.sparrow.message;

import org.zdevra.sparrow.channel.MessageChannel;
import org.zdevra.sparrow.smartproxy.SmartProxy;

import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

/**
 * Helps with {@link Message message} building.
 *
 * @author Zdenko Vrabel (zdenko.vrabel@celum.com)
 */
public class MessageBuilder<T>
{
	//-------------------------------------------------------------------------------------------------------------------
	// members
	//-------------------------------------------------------------------------------------------------------------------

	/** original message */
	private final Message<T> originalMessage;

	/** headers */
	private final Map<String, Object> headers;

	/** payload */
	private final T payload;


	//-------------------------------------------------------------------------------------------------------------------
	// constructor & factory methods
	//-------------------------------------------------------------------------------------------------------------------


	/**
	 * It's private because they are used factory methods instead of constructor.
	 */
	private MessageBuilder(T payload, Message<T> originalMessage)
	{
		this.originalMessage = originalMessage;
		this.payload = payload;
		this.headers = new HashMap<String, Object>();
	}


	/**
	 * Build new message from origin message from which is payload and headers copied.
	 *
	 * @param originalMessage
	 * @param <T>
	 *
	 * @return
	 */
	public static <T> MessageBuilder<T> fromMessage(Message<T> originalMessage)
	{
		return new MessageBuilder<T>(originalMessage.getPayload(), originalMessage);
	}


	/**
	 * Builds new message with payload
	 *
	 * @param payload
	 * @param <T>
	 *
	 * @return
	 */
	public static <T> MessageBuilder<T> withPayload(T payload)
	{
		return new MessageBuilder<T>(payload, null);
	}


	//-------------------------------------------------------------------------------------------------------------------
	// methods
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Copy all headers from original message
	 * @return
	 */
	public MessageBuilder<T> copyHeadersFromOrigin()
	{
		copyHeaders(originalMessage.getHeader());
		return this;
	}


	/**
	 * Copy headers into new message
	 *
	 * @param headers
	 * @return
	 */
	public MessageBuilder<T> copyHeaders(MessageHeader headers)
	{
		for(Map.Entry<String, Object> entry : headers.getMap().entrySet()) {
			addHeaderIfAbsent(entry.getKey(), entry.getValue());
		}
		return this;
	}


	/**
	 * Add header into new message if is absenting in builder. If the
	 * header was set before, the next adding will be ignored.
	 *
	 * @param headerKey
	 * @param value
	 * @return
	 */
	public MessageBuilder<T> addHeaderIfAbsent(String headerKey, Object value)
	{
		if (!headers.containsKey(headerKey)) {
			addHeader(headerKey, value);
		}
		return this;
	}


	/**
	 * Set correlation ID to new message from origin message's ID.
	 *
	 * @param originMessage
	 * @return
	 */
	public MessageBuilder<T> correlationIdForMessage(Message<?> originMessage)
	{
		this.addHeader(MessageHeader.CORRELATION_ID, originMessage.getHeader().getId());
		return this;
	}


	/**
	 * Set custom correlation ID to new message
	 *
	 * @param correlationId
	 * @return
	 */
	public MessageBuilder<T> correlationId(UUID correlationId)
	{
		this.addHeader(MessageHeader.CORRELATION_ID, correlationId);
		return this;
	}


	/**
	 * Set sequence number and size of sequence to new message.
	 *
	 * @return
	 */
	public MessageBuilder<T> sequenceNumber(int sequenceNum, int sequenceSize)
	{
		addHeader(MessageHeader.SEQUENCE_NUMBER, sequenceNum);
		addHeader(MessageHeader.SEQUENCE_SIZE, sequenceSize);
		return this;
	}


	/**
	 * Set the header. If the header was set before, then the old value will be
	 * overwritten by new value. It's oposite to the
	 * {@link MessageBuilder#addHeaderIfAbsent(String, Object)} method.
	 *
	 * @param headerKey
	 * @param value
	 * @return
	 */
	public MessageBuilder<T> addHeader(String headerKey, Object value)
	{
		headers.put(headerKey, value);
		return this;
	}


	/**
	 * This method exclude existing header from message. If you are copy headers
	 * from original message, sometimes you want exclude some header.
	 *
	 * @param headerKey
	 * @return
	 */
	public MessageBuilder<T> insteadOf(String headerKey)
	{
		headers.remove(headerKey);
		return this;
	}


	/**
	 * Set reply channel for message. Reply channels are used by {@link SmartProxy}.
	 *
	 * @param replyChannel
	 * @return
	 */
	public MessageBuilder<T> withReplyChannel(MessageChannel replyChannel)
	{
		addHeader(MessageHeader.REPLY_CHANNEL, replyChannel);
		return this;
	}


	/**
	 * Set return address for message. Reply channels are used by {@link SmartProxy}.
	 *
	 * @param returnAddress
	 * @return
	 */
	public MessageBuilder<T> returnAddress(UUID returnAddress)
	{
		addHeader(MessageHeader.RETURN_ADDRESS, returnAddress);
		return this;
	}


	/**
	 * builds a new message
	 *
	 * @return new instance of message
	 */
	public Message<T> build()
	{
		return new Message<T>(payload, new MessageHeader(headers));
	}


}
