/*****************************************************************************
 * Copyright 2012 Zdenko Vrabel
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package org.zdevra.sparrow.message;

import org.zdevra.sparrow.channel.MessageChannel;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

/**
 * Class represents {@link Message message} headers and holds data like timestamp etc.
 *
 * @author Zdenko Vrabel (vrabel.zdenko@gmail.com)
 */
public class MessageHeader implements Serializable
{
	//------------------------------------------------------------------------------------------------------------------
	// constants
	//------------------------------------------------------------------------------------------------------------------

	public static final String ID = "ID";
	public static final String TIMESTAMP = "TIMESTAMP";
	public static final String EXPIRATION_TIME = "EXPIRATIONTIME";
	public static final String CORRELATION_ID = "CORRELATION_ID";
	public static final String SEQUENCE_SIZE = "SEQUENCE_SIZE";
	public static final String SEQUENCE_NUMBER = "SEQUENCE_NUMBER";
	public static final String RETURN_ADDRESS = "RETURN_ADDRESS";
	public static final String REPLY_CHANNEL = "REPLY_CHANNEL";

	//------------------------------------------------------------------------------------------------------------------
	// members
	//------------------------------------------------------------------------------------------------------------------

	/** map where are headers stored */
	private final Map<String, Object> headers;

	//------------------------------------------------------------------------------------------------------------------
	// constructors
	//------------------------------------------------------------------------------------------------------------------

	/**
	 * Constructor
	 *
	 * @param hdr
	 *         map contains header's data
	 */
	public MessageHeader(Map<String, Object> hdr)
	{
		this.headers = new HashMap<String, Object>();
		putAll(hdr);
		put(ID, UUID.randomUUID());
		put(TIMESTAMP, System.currentTimeMillis());
		put(EXPIRATION_TIME, System.currentTimeMillis() + (20 * 1000));
	}

	//------------------------------------------------------------------------------------------------------------------
	// impl. methods
	//------------------------------------------------------------------------------------------------------------------

	@Override
	public String toString()
	{
		return headers.toString();
	}


	//------------------------------------------------------------------------------------------------------------------
	// methods
	//------------------------------------------------------------------------------------------------------------------

	/**
	 * Method returns a value in header
	 *
	 * @param name
	 * @return
	 */
	public Object get(String name)
	{
		Object value = this.headers.get(name);
		return value;
	}


	/**
	 * Method return all header data as map.
	 * @return
	 */
	public Map<String, Object> getMap()
	{
		return this.headers;
	}


	/**
	 * Method put the value into header
	 *
	 * @param name
	 * @param value
	 */
	public void put(String name, Object value)
	{
		this.headers.put(name, value);
	}


	/**
	 * Method put values into header
	 *
	 * @param headers
	 */
	public void putAll(Map<String, Object> headers)
	{
		this.headers.putAll(headers);
	}


	/**
	 * Method returns the unique ID of the message
	 * @return
	 */
	public UUID getId()
	{
		UUID id = (UUID) get(ID);
		return id;
	}


	/**
	 * Method returns the timestamp of the message
	 * @return
	 */
	public long getTimestamp()
	{
		long timestamp = (Long) get(TIMESTAMP);
		return timestamp;
	}


	/**
	 * Method returns the time when message expires
	 * @return
	 */
	public long getExpirationTime()
	{
		long expirationTime = (Long) get(EXPIRATION_TIME);
		return expirationTime;
	}


	/**
	 * Method returns correlation ID
	 * @return
	 */
	public UUID getCorrelationId()
	{
		UUID collerationId = (UUID) get(CORRELATION_ID);
		return collerationId;
	}


	/**
	 * Method returns sequence number
	 * @return
	 */
	public int getSequenceNumber()
	{
		int sequence_num = (Integer) get(SEQUENCE_NUMBER);
		return sequence_num;
	}


	/**
	 * Method returns size of sequence
	 * @return
	 */
	public int getSequenceSize()
	{
		int sequence_size = (Integer) get(SEQUENCE_SIZE);
		return sequence_size;
	}


	/**
	 * returns 'return address' value if it's set
	 */
	public UUID getReturnAddress()
	{
		return (UUID)get(RETURN_ADDRESS);
	}


	/**
	 * returns reply channel if it's set
	 */
	public MessageChannel getReplyChannel()
	{
		return (MessageChannel) get(REPLY_CHANNEL);
	}



}
