/*****************************************************************************
 * Copyright 2012 Zdenko Vrabel
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package org.zdevra.sparrow.smartproxy;

import org.zdevra.sparrow.message.Message;
import org.zdevra.sparrow.message.MessageBuilder;
import org.zdevra.sparrow.message.MessageHeader;
import org.zdevra.sparrow.actor.ActorHandler;
import org.zdevra.sparrow.channel.MessageChannel;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

/**
 * Smart proxy follows Smart proxy by Enterprise Integration Patterns that is solution for request-reply case.
 *
 * @author Zdenko Vrabel (zdenko.vrabel@celum.com)
 */
public class SmartProxy
{
	//-------------------------------------------------------------------------------------------------------------------
	// members
	//-------------------------------------------------------------------------------------------------------------------

	/** register of all reply channels and return addresses */
	private final ConcurrentMap<UUID, MessageChannel> replyChannels;

	/** request actor */
	private final ProxyRequestActor requestActor;

	/** reply actor */
	private final ProxyReplyActor replyActor;


	//-------------------------------------------------------------------------------------------------------------------
	// constructor
	//-------------------------------------------------------------------------------------------------------------------


	/**
	 * Constructor
	 */
	public SmartProxy()
	{
		this.replyChannels = new ConcurrentHashMap<UUID, MessageChannel>();
		this.requestActor = new ProxyRequestActor(this);
		this.replyActor = new ProxyReplyActor(this);
	}


	//-------------------------------------------------------------------------------------------------------------------
	// methods
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * returns request actor handler for smart proxy
	 */
	public ActorHandler<Object> getRequestActor()
	{
		return requestActor;
	}


	/**
	 * returns reply actor handler for smart proxy
	 */
	public ActorHandler<Object> getReplyActor()
	{
		return replyActor;
	}


	/**
	 * add reply channel to inner register and returns enriched message with 'return address'
	 *
	 * @param request
	 *          incoming request message
	 *
	 * @return new message with same header and payload enriched about 'return address'
	 */
	Message<?> registerReplyChannel(Message<?> request)
	{
		MessageChannel replyChannel = (MessageChannel) request.getHeader().getReplyChannel();
		if (replyChannel == null) {
			throw new NullPointerException("missing " + MessageHeader.REPLY_CHANNEL + " header in message " + request.toString());
		}

		UUID returnAddress = UUID.randomUUID();
		replyChannels.put(returnAddress, replyChannel);

		Message<?> msg =
				  MessageBuilder
							 .fromMessage(request)
							 .copyHeadersFromOrigin()
							 .insteadOf(MessageHeader.REPLY_CHANNEL)
							 .returnAddress(returnAddress)
							 .build();

		return msg;
	}


	/**
	 * Send message to reply channel
	 */
	MessageChannel reply(Message<?> msg)
	{
		UUID address = (UUID) msg.getHeader().getReturnAddress();
		if (address == null) {
			throw new NullPointerException("missing " + MessageHeader.RETURN_ADDRESS + " header in message " + msg.toString());
		}

		MessageChannel replyChannel = replyChannels.get(address);
		replyChannels.remove(address);
		return replyChannel;
	}


}
